@@ -1,14 +1,13 @@
 /*****************************************************************************
  *  Secure Locate                                                            *
  *  Programmed by: Kevin Lindsay                                             *
- *  Copyright (c) 1998 NetNation Communications Inc & Kevin Lindsay          *
+ *  Copyright (c) 1999 NetNation Communications Inc & Kevin Lindsay          *
  * ***************************************************************************/ 
 
 #include <stdio.h>
 #include <stdlib.h>
 
 #include "link.h"
-#include "statmalloc.h"
 
 /* Initialize 2D List */
 
@@ -17,7 +16,7 @@ init_2D_list()
 {
    dir_item *listptr;
      
-   if ((listptr = (dir_item *)statmalloc(sizeof(dir_item))) == NULL) {
+   if ((listptr = (dir_item *)malloc(sizeof(dir_item))) == NULL) {
       fprintf(stderr,"Could not allocate memory for init_2D_list!\n");
    }
    
@@ -42,7 +41,7 @@ dir_item *
 add_right(dir_item *listptr)
 {
    
-   if ((listptr->right = (dir_item *)statmalloc(sizeof(dir_item))) == NULL) {
+   if ((listptr->right = (dir_item *)malloc(sizeof(dir_item))) == NULL) {
       fprintf(stderr,"Could not allocate memory for add_right!\n");
    }
 
@@ -66,7 +65,7 @@ add_right(dir_item *listptr)
 dir_item *
 add_down(dir_item *listptr)
 {
-   if ((listptr->down = (dir_item *)statmalloc(sizeof(dir_item))) == NULL) {
+   if ((listptr->down = (dir_item *)malloc(sizeof(dir_item))) == NULL) {
       fprintf(stderr,"Could not allocate memory for add_down!\n");      
    }
    
@@ -98,12 +97,12 @@ free_right(dir_item *listptr)
       
       while (listptr->up != NULL) {
          listptr = listptr->up;
-         freemalloc(listptr->down->name);         
-         freemalloc(listptr->down);
+         free(listptr->down->name);         
+         free(listptr->down);
       }
       listptr = listptr->left;
-      freemalloc(listptr->right->name);
-      freemalloc(listptr->right);
+      free(listptr->right->name);
+      free(listptr->right);
       listptr->right = NULL;
    }
    return(listptr);
