@@ -1,6 +1,6 @@
 /*****************************************************************************
- * Secure Locate v1.4                                                        *
- * October 24, 1998                                                          *
+ * Secure Locate v1.5                                                        *
+ * May 5, 1999                                                               *
  * Programmed by: Kevin Lindsay                                              *
  * Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.          *
  *                                                                           *
@@ -9,6 +9,19 @@
  *                                                                           *
  * Patches:  Sean Mcnulty <lazy@ccf.rutgers.edu>                             *
  *                * Fixed a bug which caused the decode function to fail.    *
+ *           Ulf Betlehem <flu@iki.fi>                                       *
+ *                * multiple databases                                       *
+ *                * LOCATE_PATH environment variable support                 *
+ *                * -o, --output options                                     *
+ *           Jim Dougharty  <james.dougharty@sabre.com>                      *
+ *                * Recursive directory walk will not exit if a directory    *
+ *                * cannot be read.  May happen on some NFS directories.     *
+ *           Glen Maynard <glennm@mediaone.net>                              *
+ *                * Multiple search strings are now possible                 *
+ *                * Uses access() instead of opendir() to see if a directory *
+ *                  is readable.                                             *
+ *           R.G. Wood <rgwood@stormix.com>                                  *
+ *                * Made a Debian Package for Secure Locate                  *
  *                                                                           *
  * Report any Bugs to: klindsay@mkintraweb.com                               *
  *                                                                           *
@@ -21,6 +34,7 @@
  * Secure Locate -- search database for filenames that match patterns without 
  *                  showing files that the user using slocate does not have   
  *                  access to.                                                
+ *                                                                            
  *    Copyright (C) 1994 Free Software Foundation, Inc.                       
  *                                                                            
  *    This program is free software; you can redistribute it and/or modify    
@@ -48,17 +62,16 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <dirent.h>
-#include <netinet/in.h>
 #include <grp.h>
 #include <ctype.h>
 #include <fnmatch.h>
+#include <regex.h>
 
-#include "statmalloc.h"
 #include "link.h"
 
 /* GLOBALS */
 
-#define VERSION "Secure Locate v1.4 - Released October 24, 1998\n"
+#define VERSION "Secure Locate v1.5 - Released May 5, 1999\n"
 #define GRPFILE "/etc/group"
 #define SLOC_ESC 0x80
 #define SLOC_GRP "slocate"
@@ -71,6 +84,7 @@
 /* Printable version of WARN_SECONDS.  */
 #define WARN_MESSAGE "8 days"
 
+char **SLOCATE_PATH = NULL;
 char *SLOCATEDB = "/var/lib/slocate/slocate.db";
 char *TMPSLOCATEDB = "/var/lib/slocate/slocate.tmp";
 char *SLOCATEDB_DIR = "/var/lib/slocate/";
@@ -78,6 +92,8 @@ char *EXCLUDE_DIR;
 int EXCLUDE=0;
 int VERBOSE=0;
 int QUIET=0;
+int REGEXP=0;
+char *regexp;
 char *progname;
 char *tmp_path=NULL;
 char *cur_dir=NULL;
@@ -88,6 +104,9 @@ gid_t GID;
 int first=1;
 char slevel='1';
 unsigned short SLOC_GID=65534; /* Seems to be the default nobody gid */
+int max_queries=0;
+
+int decode_db(char *database, char *str);
 
 /* Usage */
 
@@ -95,9 +114,11 @@ void
 usage() 
 {
    printf("%s\n"
-          "Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.\n\n"
-          "search usage:   %s [-q] [-d path] [--database=path] <search string>\n"
-          "database usage: %s [-qv] [-e dir1,dir2,...] [-l level] <[-U path] [-u]>\n"
+          "Copyright (c) 1999 Kevin Lindsay & Netnation Communications Inc.\n\n"
+          "search usage:   %s [-q] [-d <path>] [--database=<path>] <search string>...\n"
+          "                %s [-r <regexp>] [--regexp=<regexp>]\n"
+          "database usage: %s [-qv] [-o <file>] [--output=<file>]\n"
+          "                %s [-e <dir1,dir2,...>] [-l <level>] <[-U <path>] [-u]>\n"
           "general usage:  %s [-Vh] [--version] [--help]\n\n"
           "   Options:\n"
           "   -u                 - Create slocate database starting at path /.\n"
@@ -107,9 +128,15 @@ usage()
           "                           0 turns security checks off. This will make\n"
           "                             searchs faster.\n"
           "                           1 turns security checks on. This is the default.\n"
-          "   -q                 - Quite mode.  Error messages are suppressed.\n"
+          "   -q                 - Quiet mode.  Error messages are suppressed.\n"
+          "   -n <num>            - Limit the amount of results shown to <num>.\n"
+          "   -r <regexp>\n"
+          "   --regexp=<regexp>  - Search the database using a basic POSIX regular\n"
+          "                        expression.\n"
+          "   -o <file>\n"
+          "   --output=<file>    - Specfies the database to create.\n"
           "   -d <path>\n"
-          "   --database=path    - Specfies path of database to search in.\n"
+          "   --database=<path>  - Specfies the path of databases to search in.\n"
           "   -h\n"
           "   --help             - Display this help.\n"
           "   -v\n"
@@ -118,13 +145,15 @@ usage()
           "   --version          - Display version.\n"
           "\n"
           "Author: Kevin Lindsay\n"
-          "Bugs: klindsay@mkintraweb.com\n"
-          "\n",VERSION,progname,progname,progname);
+          "Bugs:   klindsay@mkintraweb.com\n"
+          "FTP:    ftp.mkintraweb.com/pub/linux/slocate/\n"
+          "\n",VERSION,progname,progname,progname,progname,progname);
    
    exit(0);
 }
 
 /* My fchdir() so that I can do cleaner error checking */
+
 int
 MYfchdir(int fd, char *filepath, char *func)
 {   
@@ -164,48 +193,57 @@ get_short (FILE *fp)
    return (x << 8) | (fgetc (fp) & 0xff);
 }
 
-/* Parse Path */
+/* Parse Output Path */
+
 void
-parse_path(char *path)
+parse_create_path(char *path)
 {
-   int ptrlen;
-   int grow;
    struct stat statres;
-   char *ptr;
 
-   ptrlen = strlen(path);   
-   if (stat(path,&statres) == -1) {
-      if (!QUIET)
-          fprintf(stderr,"%s: error accessing path: %s : %s\n",progname,path,strerror(errno));
-      exit(1);
-   }
-   
+   if (!path || strlen(path) == 0) return;
+
+   stat(path, &statres);
    if (S_ISDIR(statres.st_mode)) {
-      if (!QUIET)
-          fprintf(stderr,"%s: %s is a directory\n",progname,path);
-      exit(1);
+   if (!QUIET)
+        fprintf(stderr,"%s: %s is a directory\n", progname,path);
+   exit(1);
    }
-
-   grow=ptrlen+1;
-   SLOCATEDB = malloc(grow);
+   SLOCATEDB = malloc(strlen(path)+1);
    SLOCATEDB[0] = '\0';
-   strcat(SLOCATEDB,path);
-   
-   if ((ptr = strrchr(path,'/')) != NULL) {
-      ptr++;
-      SLOCATEDB_DIR = malloc((ptr-path)+1);
-      SLOCATEDB_DIR[0] = '\0';
-      strncat(SLOCATEDB_DIR,path,ptr-path);
-   } else {
-      SLOCATEDB_DIR = malloc(8);
-      strcat(SLOCATEDB_DIR,".\0");
-   }
+   strcat(SLOCATEDB, path);
 
-   TMPSLOCATEDB = malloc(strlen(SLOCATEDB)+5);
+   TMPSLOCATEDB = malloc(strlen(path)+5);
    TMPSLOCATEDB[0] = '\0';
-   strcat(TMPSLOCATEDB,SLOCATEDB);
-   strcat(TMPSLOCATEDB,".tmp");
+   strcat(TMPSLOCATEDB, path);
+   strcat(TMPSLOCATEDB, ".tmp");
+}
 
+/* Parse Path */
+
+void
+parse_decode_path(char *path)
+{
+   char *pathcopy;
+   char *part;
+   int i;
+
+   if (!path || strlen(path) == 0) return;
+
+   i = 1;
+   part = path;
+   while ((part = strchr(part+1, ':'))) i++;
+   SLOCATE_PATH = malloc(i * sizeof(char *));
+
+   i = 0;
+   pathcopy = malloc(strlen(path)+1);
+   pathcopy[0] = '\0';
+   strcat(pathcopy, path);
+   part = strtok(pathcopy, ":");
+   while(part) {
+   SLOCATE_PATH[i++] = part;
+   part = strtok(NULL, ":");
+   }
+   SLOCATE_PATH[i] = NULL;
 }
 
 /* Parse Dash */
@@ -214,6 +252,9 @@ void
 parse_dash(char *option)
 {
    char *ptr;
+   int i=0;
+   char *database;
+   int res=1;
    
    for (ptr = option;*ptr != 0 && *ptr != '=' ; ptr++)
        *ptr = toupper(*ptr);
@@ -231,10 +272,24 @@ parse_dash(char *option)
    
    *ptr = '\0';
    ptr++;
+
+   if (!strcmp(option,"OUTPUT")) {
+      parse_create_path(ptr);
+   }
+
    if (!strcmp(option,"DATABASE")) {
-      parse_path(ptr);
+      parse_decode_path(ptr);
    }
    
+   if (!strcmp(option,"REGEXP")) {
+      REGEXP = 1;
+      regexp = malloc(strlen(ptr)+1);
+      *regexp = 0;
+      strcat(regexp,ptr);
+      while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, ptr);
+      exit(!res);
+   }
+
 }
 
 
@@ -329,7 +384,7 @@ match_exclude(char *path,char *filedir) {
    int res=0;
    char *newstr;
    
-   newstr = statmalloc(strlen(path)+strlen(filedir)+3);
+   newstr = malloc(strlen(path)+strlen(filedir)+3);
    newstr[0] = '\0';
    strcat(newstr,"*");
    strcat(newstr,path);
@@ -339,7 +394,7 @@ match_exclude(char *path,char *filedir) {
    if (strstr(EXCLUDE_DIR,newstr) != NULL)       
        res=1;
    
-   freemalloc(newstr);
+   free(newstr);
    
    return res;
 }
@@ -388,7 +443,7 @@ load_dir(int dir_fd, dir_item *listptr)
          
          /* If directory is emtpy then add a space to listptr */         
          if (filename == NULL) {
-            listptr->name = statmalloc(1);
+            listptr->name = malloc(1);
             listptr->name[0] = '\0';
             listptr->empty = 1;
             listptr->type = 0;
@@ -404,9 +459,8 @@ load_dir(int dir_fd, dir_item *listptr)
          listptr = listptr->down;
       }
       
-      
       /* Add information from filenames->d_name to listptr */
-      listptr->name = statmalloc(strlen(filename->d_name)+1);
+      listptr->name = malloc(strlen(filename->d_name)+1);
       listptr->name[0] = '\0';
       strcat(listptr->name,filename->d_name);
       listptr->name_len = strlen(listptr->name);      
@@ -612,7 +666,7 @@ create_db(char *dirstr)
    }
 
    
-   listptr->name = statmalloc(strlen(dirstr)+1);
+   listptr->name = malloc(strlen(dirstr)+1);
    listptr->name[0] = '\0';
    strcat(listptr->name,dirstr);
    listptr->st_uid = statres.st_uid;
@@ -724,8 +778,12 @@ create_db(char *dirstr)
             
             if ((dir_fd = open(".",O_RDONLY)) == -1) {
                if (!QUIET)
-                   fprintf(stderr,"%s: open: CDB_5: %s\n",progname,strerror(errno));
-               return(1);
+                 fprintf(stderr,"%s: open: CDB_5: %s: %s%s\n",progname,strerror(errno),
+                         get_cur_dir(listptr,dirstr),listptr->name);
+               //   return(1);  Don't bail just because we can't read a directory.
+               // This might happen even as root on some NFS partitions.
+               load=0;
+               list_down=0;          
             }
          }
                   
@@ -742,11 +800,10 @@ create_db(char *dirstr)
 /* Decode Database */
 
 int
-decode_db(char *str)
+decode_db(char *database, char *str)
 {
    int res = 1;
    FILE *fd;
-   DIR *tmpfd;
    short code_num;
    int pathlen=0;
    register char ch;
@@ -763,19 +820,30 @@ decode_db(char *str)
    time_t now;
    char *chk1;
    char tmpch;
+   regex_t *preg=NULL;
+   char *errbuf=NULL;
+   int nmatch=32;
+   regmatch_t pmatch[32];
+   int cur_queries = 0;
    
-   if ((fd = fopen(SLOCATEDB,"r")) == NULL) {
+   if ((fd = fopen(database,"r")) == NULL) {
       if (!QUIET)
-          fprintf(stderr,"%s: decode_db(): open: (%d) %s\n",progname,errno,strerror(errno));
+          fprintf(stderr,"%s: decode_db(): %s: %s\n",progname,database,strerror(errno));
       return(0);
    }
 
-   lstat(SLOCATEDB,&statres);
+   lstat(database,&statres);
+
+   if (S_ISDIR(statres.st_mode)) {
+   if (!QUIET)
+        fprintf(stderr,"%s: decode_db(): %s is a directory\n",progname,database);
+   return(0);
+   }
 
    time(&now);
    if (now - statres.st_mtime > WARN_SECONDS)
    {
-      fprintf(stdout,"%s: warning: database %s' is more than %s old\n",progname,SLOCATEDB,WARN_MESSAGE);
+      fprintf(stdout,"%s: warning: database %s' is more than %s old\n",progname,database,WARN_MESSAGE);
    }   
    
    slevel = getc(fd);
@@ -832,7 +900,19 @@ decode_db(char *str)
       
       printit=0;
       
-      if ((globflag && !fnmatch(str,codedpath,0)) || strstr(codedpath,str) != NULL) {
+      if (REGEXP) {
+         globflag=0;
+         preg = malloc(sizeof(regex_t));
+         if ((res = regcomp(preg,regexp,0)) != 0) {
+            regerror(res, preg, errbuf,1024);
+            if (!QUIET)
+              fprintf(stdout,"error: %s: regular expression: %s\n",progname,errbuf);
+            exit(1);
+         }         
+         
+      }
+
+      if ((globflag && !fnmatch(str,codedpath,0)) || (!REGEXP && strstr(codedpath,str) != NULL) || (REGEXP && regexec(preg,codedpath,nmatch,pmatch,0) == 0)) {
          if (slevel == '1') {
             if (UID == 0 || lstat(codedpath,&statres) == 0) {
                printit=1;
@@ -842,11 +922,8 @@ decode_db(char *str)
                   tmpch=*chk1;
                   *chk1='\0';
 
-                  if ((tmpfd = opendir(codedpath)) == NULL)
-                      printit=0;
-                  else
-                      closedir(tmpfd);
-                  
+                  if (access(codedpath, R_OK) == -1)
+                    printit=0;
                   *chk1=tmpch;
                }
             }
@@ -854,13 +931,20 @@ decode_db(char *str)
              printit=1;
       }
       
-      if (printit)
+      if (REGEXP)
+        regfree(preg);
+      
+      if (printit) {
+         cur_queries += 1;
          printf("%s\n",codedpath);
+         if (max_queries > 0 && cur_queries == max_queries)
+           exit(0);
+      }
    }      
    
    fclose(fd);
    
-   return(res);
+   return(1);
 }
 
 /* Main Function */
@@ -876,20 +960,29 @@ main(int args, char **argv)
    int SPECDIR=0;
    int ROOTDIR=0;   
    char *spec_dir=NULL;
-   int optlen;
+   char *database;
+   int i=0;
+
 
    progname = ((p = strrchr(argv[0],'/')) ? p+1 : *argv);   
    
-   if (args < 2) 
-       usage();
-
+   if (!strcmp(progname,"updatedb"))
+     ROOTDIR=1;
+   
+   if (args < 2 && !ROOTDIR)
+     usage();
+   
    if (!check_dir())
-       return(1);   
+     return(1);
    
    UID = getuid();        
    GID = getgid();
-   
-   while ((ch = getopt(args,argv,"VvuhqU:e:l:d:-:")) != EOF) {
+
+   parse_create_path(SLOCATEDB);
+   parse_decode_path(SLOCATEDB);
+   parse_decode_path(getenv("LOCATE_PATH"));
+
+   while ((ch = getopt(args,argv,"VvuhqU:r:o:e:l:d:-:n:")) != EOF) {
       switch(ch) {
       case 'h':
          usage();
@@ -907,30 +1000,50 @@ main(int args, char **argv)
       case 'e':
          EXCLUDE_DIR = parse_exclude(optarg);
          break;
-      case 'l':
+       case 'l':
          slevel = optarg[0];
          if (slevel != '0' && slevel != '1') {
             if (!QUIET)
-                fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
+              fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
             exit(1);
          }
          break;
       case 'u':
-         ROOTDIR=1;         
+         ROOTDIR=1;
          break;
       case 'U':
+         ROOTDIR=0;
          SPECDIR=1;
-         optlen=strlen(optarg);
-         spec_dir = malloc(optlen+1);
-         memmove(spec_dir,optarg,optlen);
-         spec_dir[optlen] = '\0';         
+         spec_dir = malloc(strlen(optarg)+1);
+         *spec_dir = 0;         
+         strcat(spec_dir,optarg);
+         break;
+       case 'r':
+         REGEXP = 1;
+         regexp = malloc(strlen(optarg)+1);
+         *regexp = 0;
+         strcat(regexp,optarg);
+         while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, optarg);
+         return(!res);
+         break;
+      case 'o':
+         parse_create_path(optarg);
          break;
       case 'd':
-         parse_path(optarg);
+         parse_decode_path(optarg);
          break;
       case '-':
          parse_dash(optarg);
          break;
+       case 'n':
+         for (i=0; i < strlen(optarg); i+=1) {
+            if (!isdigit(optarg[i])) {
+               fprintf(stderr,"%s: ERROR: Invalid argument for option -n\n",progname);
+               exit(1);
+            }
+         }
+         max_queries = atoi(optarg);
+         break;
       default:
          return(1);
          break;
@@ -942,16 +1055,18 @@ main(int args, char **argv)
    
    if (SPECDIR) {
       res = create_db(spec_dir);
-   }
-   else if (ROOTDIR)
+   } else if (ROOTDIR)
       res = create_db((char *)NULL);
    else if (argv[optind]) {
-      res = decode_db(argv[optind]);
+      int j = optind;
+      int i = 0;
+      
+      while (j < args) {
+         while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, argv[j++]);
+         i = 0;
+      }
    } else
        usage();
 
-   if (!res)
-       return(1);
-   
-   return(0);
+   return(!res);
 }
